#!/usr/bin/env python

"""
Preparation of run-time attribute lookup tables.

Copyright (C) 2007, 2008, 2009, 2010 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from micropython.common import *
from micropython.data import *

class List:

    """
    A displaced list containing attribute details and an index mapping names to
    offsets in the list. This is the optimised form of the table initially
    constructed to record object attributes.
    """

    def __init__(self, names):
        self.names = names
        self.displaced = []
        self.offsets = {}
        self.offsets_used = {}

        self.names_index = {}
        for i, name in enumerate(self.names):
            self.names_index[name] = i

    def __len__(self):
        return len(self.displaced)

    def __getitem__(self, i):
        return self.displaced[i]

    def get_code(self, name):

        "Return the code/offset of the given 'name'."

        return self.offsets.get(name)

    def get_name(self, code):

        "Return the name for the given 'code' or offset."

        return self.offsets_used[code]

    # Simulation methods.

    def access(self, objname, attrname):
        obj_offset = self.offsets.get(objname)
        attr_index = self.names_index.get(attrname)
        if obj_offset is None or attr_index is None:
            return None

        element = self.displaced[obj_offset + attr_index]
        if element is not None:
            index, details = element
            if index == attr_index:
                return details

        return None

    def dir(self, objname):
        attributes = {}
        for name in self.names:
            attr = self.access(objname, name)
            if attr is not None:
                attributes[name] = attr
        return attributes

    # Update methods.

    def add(self, objname, attributes):

        "Add an entry for 'objname' with the given 'attributes'."

        len_displaced = len(self.displaced)
        len_attributes = len(attributes)

        # Try to fit the row into the list, starting from the beginning and
        # skipping places where an existing row has been fitted.

        for offset in xrange(0, len_displaced):
            if self.offsets_used.has_key(offset):
                continue
            if self._fit_row(offset, attributes, len_displaced):
                self.offsets_used[offset] = objname
                self.offsets[objname] = offset
                self._add_row(offset, attributes, len_displaced)
                break

        # Where a row could not be fitted within the list, add it to the
        # end.

        else:
            self.offsets_used[len_displaced] = objname
            self.offsets[objname] = len_displaced
            self._add_row(len_displaced, attributes, len_displaced)

    def _fit_row(self, offset, attributes, len_displaced):

        """
        Fit, at the given 'offset', the row of 'attributes' in the displaced
        list having length 'len_displaced'.
        Return a true value if this succeeded.
        """

        for i, attr in enumerate(attributes):
            if i + offset >= len_displaced:
                break
            element = self.displaced[i + offset]
            if attr is not None and element is not None:
                return 0
        return 1

    def _add_row(self, offset, attributes, len_displaced):

        """
        Add, at the given 'offset', the row of 'attributes' in the displaced
        list having length 'len_displaced'.
        """

        # Extend the list if necessary.

        for i in xrange(0, offset + len(attributes) - len_displaced):
            self.displaced.append(None)

        # Record the attribute details in the list.

        for i, attr in enumerate(attributes):
            if attr is not None:

                # Each entry is of the form (attribute number, attribute).

                self.displaced[offset+i] = i, attr

    # Image production.

    def as_raw(self):

        "Return the raw contents of the table as a list of values."

        result = []
        for entry in self.displaced:
            if entry is None:
                result.append(None)
            else:
                result.append(self.entry_as_raw(entry))

        return result

class ObjectList(List):

    "An object list."

    def entry_as_raw(self, entry):
        attr_index, attr = entry

        # Support descendants.

        if isinstance(attr, Class):
            return (attr_index, None, None)

        if attr.parent is not None:
            location = attr.parent.location or 0
        else:
            location = 0
        if attr.position is not None:
            position = attr.position + location + 1 # skip structure header
        else:
            position = None # NOTE: Should fix unpositioned attributes.

        # Attribute index/code, attribute type, location/position.

        return (attr_index, attr.is_static_attribute(), position)

class ParameterList(List):

    "A parameter list."

    def entry_as_raw(self, entry):
        # param_index, position = entry
        return entry

class Table:

    "A lookup table."

    TableError = TableError
    list_class = None # overridden

    def __init__(self):
        self.attributes = set()
        self.table = {}
        self.objnames = []
        self.names = []
        self.displaced_list = None
        self.raw = None

    def access(self, objname, attrname):

        "Return the attribute entry for the given 'objname' and 'attrname'."

        try:
            entry = self.table[objname]
        except KeyError:
            raise TableError, "Name %r is not registered as an object in the table." % objname

        try:
            return entry[attrname]
        except KeyError:
            raise TableError, "Name %r is not registered as an attribute in the table." % attrname

    def add(self, objname, attributes):

        "For the given 'objname' add the given 'attributes' to the table."

        self.table[objname] = attributes
        for name, origin in attributes.items():
            self.attributes.add(name)

        if self.displaced_list is not None:
            self.displaced_list.add(objname, self.matrix_row(attributes))

    def object_names(self):

        "Return the object names used in the table."

        self.objnames = self.objnames or list(self.table.keys())
        return self.objnames

    def attribute_names(self):

        "Return the attribute names used in the table."

        self.names = self.names or list(self.attributes)
        return self.names

    def get_index(self, name):

        "Return the index of the given 'name'."

        try:
            return self.attribute_names().index(name)
        except ValueError:
            raise TableError, "Name %r is not registered as an attribute in the table." % name

    def as_matrix(self):

        """
        Return the table as a matrix mapping object names to lists of attributes
        arranged in the order of the recorded attribute names.
        """

        matrix = {}
        for objname, attributes in self.table.items():
            matrix[objname] = self.matrix_row(attributes)
        return matrix

    def matrix_row(self, attributes):

        "Return a matrix row for the given 'attributes'."

        row = []
        for name in self.attribute_names():
            row.append(attributes.get(name))
        return row

    def all_attribute_positions(self, name):

        """
        Return a list of positions for the attribute with the given 'name' from
        all known objects.
        """

        all = set()
        for attributes in self.table.values():
            if attributes.has_key(name):
                all.add(attributes[name])
        return all

    def all_possible_objects(self, names):

        """
        Return a list of object names supporting the given attribute 'names'.
        """

        possible = []
        for objname, attributes in self.table.items():
            found = 1
            for name in names:
                if not attributes.has_key(name):
                    found = 0
                    break
            if found:
                possible.append(objname)
        return possible

    def any_possible_objects(self, names):

        """
        Return a list of object names supporting any of the given attribute 'names'.
        """

        possible = []
        for objname, attributes in self.table.items():
            found = 0
            for name in names:
                if attributes.has_key(name):
                    found = 1
                    break
            if found:
                possible.append(objname)
        return possible

    def as_list(self):

        """
        Return a displaced list object encoding the table in a more compact form
        than that provided by the matrix representation. This list will be
        updated with new entries if added to this object using the 'add' method.
        """

        if self.displaced_list is None:
            self.displaced_list = self.list_class(self.attribute_names())

            # Visit each row of the matrix.

            matrix_by_usage = []
            size = len(self.attributes)

            # Add rows in descending order of utilisation.

            for objname, attributes in self.as_matrix().items():
                matrix_by_usage.append((size - attributes.count(None), objname, attributes))

            matrix_by_usage.sort()
            matrix_by_usage.reverse()

            for usage, objname, attributes in matrix_by_usage:
                self.displaced_list.add(objname, attributes)

        return self.displaced_list

    def as_raw(self):

        "Return the raw contents of the table as a list of values."

        if self.raw is None:
            self.raw = self.as_list().as_raw()
        return self.raw

class ObjectTable(Table):

    "An object table."

    list_class = ObjectList

    def all_possible_objects_plus_status(self, names):

        """
        Return all objects supporting attributes with the given 'names' plus
        whether such attributes can be exclusively static or may belong to
        instances. A list of tuples of the form (object name, is_static) is thus
        returned.
        """

        possible = []
        for objname in self.all_possible_objects(names):
            attributes = self.table[objname]

            is_static = 0
            is_instance = 0

            for name in names:
                if attributes[name].is_static_attribute():
                    is_static = 1
                else:
                    is_instance = 1

            if is_static and not is_instance:
                is_static = 1
            else:
                is_static = 0

            possible.append((objname, is_static))

        return possible

class ParameterTable(Table):

    "A parameter table."

    list_class = ParameterList

# vim: tabstop=4 expandtab shiftwidth=4
